#
#   Created By Ryan McDermott
#   Created on 5/4/2022
#

import datetime
import uuid
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Wrapper:
    '''
        Wrapper for Google Cloud Storage python SDK
    '''
    __google_cloud_storage_client = None

    def __init__(self):
        self.__google_cloud_storage_client = GcpClients.instance(
        ).get_storage_client()

    def find_blobs_with_metadata(self, bucket, key, value, path_prefix=''):
        '''
            Finds every file with with the metadata matching the key and value pairs
            provided. Can also set a prefix to narrow down the available searches
            Args:
                key (string): The metadata key
                value (string): the metadata value
                path_prefix (string): blob path prefix
            Returns:
                list: list containing all blobs with mathcing metadata
        '''
        blobs = []
        for blob in self.__google_cloud_storage_client.list_blobs(
                bucket, prefix=path_prefix):
            if blob.metadata is not None and key in blob.metadata and blob.metadata[
                    key] == value:
                blobs.append(blob.name)

        return blobs

    def check_if_bucket_exists(self, bucket_name):
        '''
            checks if the bucket name passed exists
            Args:
                bucket_name (string): the name of the bucket
            Returns:
                boolean: True if the bucket exists, False if otherwise
        '''
        try:
            self.__google_cloud_storage_client.get_bucket(bucket_name)
            return True
        except Exception as e:
            print("bucket " + bucket_name + " does not exist")
            return False

    def upload_file_to_gcs_bucket(self, mrn, file_path, prefix, client_uuid):
        '''
            uploads file to specified folder in data lake - prefix determines the folder
            args:
                mrn: str - user input mrn
                file_path: str - local file path of file to upload
                prefix: str - folder on data lake where we want file to be placed
            return:
                upload_result: boolean - true if file upload successful, false if failure encountered
        '''
        if 'gs://' in file_path:
            bucket_name = file_path.replace("gs://", '').split('/')[0]
            gcs_file_path = file_path.replace("gs://" + bucket_name + "/", '')
            blob_name = prefix + "/" + mrn + "_" + str(
                uuid.uuid4()) + "_" + file_path.split('/')[-1]
            source_bucket = self.__google_cloud_storage_client.bucket(
                bucket_name)
            destination_bucket = self.__google_cloud_storage_client.bucket(
                client_uuid)
            blob = source_bucket.blob(gcs_file_path)
            blob_copy = source_bucket.copy_blob(blob,
                                                destination_bucket,
                                                new_name=blob_name)

            if blob_copy.exists():
                return blob_copy
            return False
        else:
            file_path = file_path.replace("\\", "/")
            file_name = file_path.split("/")[-1]
            blob_name = prefix + "/" + mrn + "_" + str(
                uuid.uuid4()) + "_" + file_name
            bucket = self.__google_cloud_storage_client.bucket(client_uuid)
            blob = bucket.blob(blob_name)
            upload_result = self.__upload_blob_from_local(blob, file_path)

            if not upload_result:
                return upload_result
            return blob

    def __upload_blob_from_local(self, blob, file_path):
        '''
            applies gcs api to upload a file from local to blob on gcs
            args:
                blob: gcs blob object
                file_path: local file path of file to be uploaded
            return:
                boolean: true is success and false if failure encountered
        '''
        try:
            blob.upload_from_filename(file_path)
            return True
        except Exception as e:
            Helpers.print_error("Could not connect to GCS client " + str(e))
            return False



#
#   Created by Ryan McDermott
#   Created on 3/16/2022
#

from google.api_core.exceptions import NotFound

from google.cloud import compute_v1
from google.cloud.compute_v1.types import Items
from google.cloud.compute_v1.types import AttachedDisk
from google.cloud.compute_v1.types import ServiceAccount
from google.cloud.compute_v1.types import AccessConfig
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Wrapper:
    ''' Class that wraps the Compute Engine API Client '''
    __compute_engine_instance_client = None
    __compute_engine_operation_client = None

    def __init__(self):
        self.__compute_engine_instance_client = GcpClients.instance(
        ).get_compute_engine_instance_client()
        self.__compute_engine_operation_client = GcpClients.instance(
        ).get_compute_engine_operation_client()

    def check_if_vm_exists(self, vm_name, zone_id, project_id):
        '''
            Function that checks if a VM exists
            Args:
                vm_name (string): The name of the vm
            Returns:
                boolean: True if the compute engine exists False if otherwise
        '''
        try:
            self.__compute_engine_instance_client.get(project=project_id,
                                                      zone=zone_id,
                                                      instance=vm_name)
            return True
        except NotFound:
            return False

    def create_disk(self, source_image, disk_size_gb, disk_type, auto_delete,
                    boot, mode):
        ''' 
            Creates a disk instance 
            Args:
                source_image (string): The source image of the disk 
                    (projects/{image-family}/global/images/{image-name})
                disk_size_gb (int): The disk size in GB
                disk_type (string): The disk type (zones/{zone_id}/diskTypes/{disk_type})
                auto_delete (boolean): The flag to set auto delete after instance delete
                boot (boolean): flag to set if this is a boot disk
                mode (string): The mods to set
            Returns:
                google.cloud.compute_v1.types.AttachedDisk: The created disk
        '''
        attached_disk = AttachedDisk()
        initialize_params = compute_v1.AttachedDiskInitializeParams()
        initialize_params.source_image = source_image
        initialize_params.disk_size_gb = disk_size_gb
        initialize_params.disk_type = disk_type
        attached_disk.initialize_params = initialize_params
        attached_disk.auto_delete = auto_delete
        attached_disk.boot = boot
        attached_disk.mode = mode

        return attached_disk

    def create_service_account_instance_with_scopes(self, service_account_email,
                                                    scopes):
        ''' 
            Creates a service account instance with scopes to be used for the VM
            Args:
                service_account_email (string): the service account to authenticate to services
                scopes (list): list of scopes that the VM can use
            Returns:
                google.cloud.compute_v1.types.ServiceAccount: The created resource
        '''
        service_account = ServiceAccount()
        service_account.email = service_account_email
        service_account.scopes = scopes

        # service account must be iterable for compute instance creations
        return service_account

    def create_compute_instance(self, vm_name, metadata, zone_id, machine_type,
                                disks, network_interfaces, service_accounts):
        '''
            Creates a compute engine instance 
            Args:
                vm_name (string): The name of the vm
                metadata (Sequence[Tuple[str, str]]): the vm metadata
                machine_type (string): The machine type
                disks (list[google.cloud.compute_v1.types.AttachedDisk]): The attached disks
                network_interfaces (compute_v1.NetworkInterface): The network interface
                service_account (list[compute_v1.ServiceAccount]): The Service Account
            Returns:
                Instance: The instance we wish to create
        '''
        instance = compute_v1.Instance()
        instance.name = vm_name
        instance.metadata = metadata
        instance.disks = disks
        instance.machine_type = 'zones/' + zone_id + '/machineTypes/' + machine_type
        instance.network_interfaces = network_interfaces
        instance.service_accounts = service_accounts

        return instance

    def create_insert_instance_request(self, instance, zone_id, project_id):
        '''
            Creates a request instance object for creating a compute engine instance
            Args:
                instance: The instance object we wish to create
            Returns
                compute_v1.InsertInstanceRequest: The request instance we wish to create
        '''
        request = compute_v1.InsertInstanceRequest()
        request.zone = zone_id
        request.project = project_id
        request.instance_resource = instance

        return request

    def create_compute_engine_instance_with_request(self, request):
        ''' 
            Creates a compute engine instance with a 
            compute_v1.InsertInstanceRequest instance 
            Args:
                request (compute_v1.InsertInstanceRequest): The request we want to create
            Returns:
                compute_v1.operation: The created operation
        '''
        try:
            return self.__compute_engine_instance_client.\
                insert(request=request)
        except Exception as exception:
            Helpers.print_error('Could not create Compute Engine Instance ' +
                                str(exception))
            return None

    def stop_compute_engine_instance(self, instance_name, zone_id, project_id):
        '''
            Stops a compute engine Instance
            Args:
                instance_name (string): The name of the instance
                zone_id (string): The id of the zone
                project_id (string): The id of the project
            Returns:
                compute_v1.operation|None: The created operation

        '''
        try:
            return self.__compute_engine_instance_client.\
            stop(project=project_id, zone=zone_id, instance=instance_name)
        except Exception as e:
            Helpers.print_error('Failed To Stop instance ' + str(e))
            return None

    def create_default_network_interface(self):
        ''' 
            Creates an empty network interface object and returns it
            Args:
                None
            Returns:
                compute_v1.NetworkInterface
        '''
        network_interface = compute_v1.NetworkInterface()
        network_interface.network = 'global/networks/default'
        network_interface.access_configs = [AccessConfig()]

        return network_interface

    def wait_for_operation(self, operation, zone_id, project_id):
        ''' 
            Wait for the operation to be completed before returning True 
            Args:   
                operation (compute_v1.operation): the operation we are waiting for
                zone_id (string): The zone id of the operation
                project_id (string): The project id for the operation
            Returns:
                boolean: True if the operation was successful, False if otherwise
        '''
        while operation.status != compute_v1.Operation.Status.DONE:
            operation = self.__compute_engine_operation_client.wait(
                operation=operation.name, zone=zone_id, project=project_id)
        if operation.error:
            Helpers.print_error("Error during creation: " +
                                str(operation.error))
            return False
        if operation.warnings:
            Helpers.print_error("Warning during creation: " +
                                str(operation.warnings))
            return False

        print(operation.name + ': Operation has finished')
        return True

    def create_then_stop_compute_instance(self, request, vm_name, zone_id,
                                          project_id):
        ''' 
            Function that creates then stops a VM instance
            Args:
                request (compute_v1.InsertInstanceRequest): The request to make a VM instance
                vm_name (string): The name of the VM we wish to create and stop
                zone_id (string): The Id of the zone
                project_id (string): The id of the project
            Returns:
                boolean: True if the instance was created and stopped, 
                    False if otherwise
        '''

        if not self.__create_compute_instance(request, zone_id, project_id):
            return False

        return self.__stop_compute_instance(vm_name, zone_id, project_id)

    def __create_compute_instance(self, request, zone_id, project_id):
        '''
            Performs logic and checks to create a compute instance
            Args:
                request (compute_v1.InsertInstanceRequest): The request to make a VM instance
                zone_id (string): The Id of the zone
                project_id (string): The id of the project
            Returns:
                boolean: True if the instance was created, 
                    False if otherwise
        '''
        create_operation               = self.\
            create_compute_engine_instance_with_request(request)

        if create_operation is None:
            return False

        if not self.\
            wait_for_operation(create_operation, zone_id, project_id):
            Helpers.print_error(
                'Operation to create VM instance did not complete')
            return False

        return True

    def __stop_compute_instance(self, vm_name, zone_id, project_id):
        '''
            Performs logic and checks to stop a compute instance
            Args:
                vm_name (string): The name of the VM we wish to create and stop
                zone_id (string): The Id of the zone
                project_id (string): The id of the project
            Returns:
                boolean: True if the instance was created, 
                    False if otherwise
        '''
        stop_operation               = self.\
            stop_compute_engine_instance(vm_name, zone_id, project_id)

        if stop_operation is None:
            return False

        if not self.\
            wait_for_operation(stop_operation, zone_id, project_id):
            Helpers.print_error(
                'Operation to stop VM instance did not complete')
            return False

        return True

    def create_metadata_item(self, key, value):
        '''
            Creates a metadata item for the backup compute engine
            Args:
                key (string): The key to the metadata
                value (string): The value to the key / value pair for the metadata
            Returns:
                google.cloud.compute_v1.types.Items: metadata item for client id
        '''
        item = Items()
        item.key = key
        item.value = value

        return item

    def create_metadata_item_from_file(self, key, file_path):
        '''
            Creates a metadata item from a file for the backup compute engine
            Args:
                key (string): The key to the metadata
                file_path (string): The path to the file we want to set as metadata
            Returns:
                google.cloud.compute_v1.types.Items: metadata item for client id
        '''
        item = Items()
        item.key = key
        item.value = self.__get_contents_from_file(file_path)

        return item

    def __get_contents_from_file(self, file_path):
        ''' 
            reads the contents of a file from the filepath and returns them
            Args:
                file_path (string): The path to the file we want to read
            Returns:
                string: The contents of the file
        '''
        with open(file_path, "r", encoding='utf-8') as file:
            return file.read()



#
#   Created by Ryan McDermott
#   Created on 3/11/2022
#

import datetime

from google.cloud.bigquery_storage_v1 import types
from google.cloud.bigquery_storage_v1 import writer
from google.protobuf import descriptor_pb2

from indxdatalaketools.ClientTools.Changelog import changelog_schema_pb2
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClientWrappers import BigQuery
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Client:
    ''' Class that inserts data into the changelog table '''
    client_uuid = ''
    table_fields = [
        "CLIENT_ID", "BUCKET_ID", "FILE_PATH", "FILE_NAME", "ACTION",
        "ACTION_AT", "REPLICATED", "REPLICATED_AT"
    ]

    __big_query_wrapper = None
    __google_big_query_storage_client = None
    __proto_buf_dictionary = None

    def __init__(self, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google api clients used to insert items to the Patients table
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid
        logging_client = GcpClients.instance().get_logging_client()
        self.__changelog_insert_logger = logging_client.logger(
            client_uuid + '-changelog-logger')
        self.__google_big_query_storage_client = GcpClients.instance(
        ).get_bigquery_storage_v1_client()
        self.__big_query_wrapper = BigQuery.Wrapper()

        self.__create_proto_buf_dictionary()

    def __create_proto_buf_dictionary(self):
        '''
            Function that creates the protobuf ditionary to use to populate 
            its attributes easily
        '''
        self.__proto_buf_dictionary = {
            "CLIENT_ID": self.__insert_client_id,
            "BUCKET_ID": self.__insert_bucket_id,
            "FILE_PATH": self.__insert_file_path,
            "FILE_NAME": self.__insert_file_name,
            "ACTION": self.__insert_action,
            "ACTION_AT": self.__insert_action_at,
            "REPLICATED": self.__insert_replicated,
            "REPLICATED_AT": self.__insert_replicated_at
        }

    def insert_to_changelog_table(self, blob, action):
        ''' 
            Inserts information from an uploaded blob to the changelog table
            Args:
                blob (google.cloud.storage.Blob): The blob that was uploaded
            Returns:
                boolean: True if the data was inserted into the changelog table
        '''
        blob_data_dict = self.__create_upload_dict(blob, action)
        dataset_id = 'SNAPSHOTS_' + self.__transform_client_id()
        today = self.__get_date_today()

        return self.__insert_data_to_changelog(dataset_id, today,
                                               blob_data_dict)

    def __create_upload_dict(self, blob, action):
        '''
            Creates a dictionary with all of the data for the table
            Args:
                blob   (google.cloud.storage.Blob): The blob that was 
                    created, updated, or deleted
                action (string): the action being taken
            Returns:
                dict: The dictionary containing all of the data
        '''
        time_now = self.__get_datetime_now()

        client_id = self.client_uuid
        bucket_id = self.client_uuid
        file_path = blob.metadata['FILE_PATH']
        file_name = str(blob.name).split('/')[-1]
        action_at = time_now
        replicated = False

        return {
            'CLIENT_ID': client_id,
            'BUCKET_ID': bucket_id,
            'FILE_PATH': file_path,
            'FILE_NAME': file_name,
            'ACTION': action,
            'ACTION_AT': action_at,
            'REPLICATED': replicated
        }

    def __get_datetime_now(self):
        '''
            Gets the current datetime for now and returns it
            Returns:
                datetime: The datetime of now
        '''
        now = datetime.datetime.utcnow()
        time_now_string = now.strftime('%Y-%m-%d %H:%M:%S')

        return time_now_string

    def __transform_client_id(self):
        '''
            Transforms the client id to contain no dashes and have all
            letters uppercase:
            Args:
                None
            Returns:
                string: transformed string
        '''
        client_id = self.client_uuid
        client_id = client_id.replace('-', '')
        client_id = client_id.upper()

        return client_id

    def __get_date_today(self):
        '''
            creates a date string in YYYYMMDD format by using the date today 
            Returns:
                string: Date today in YYYYMMDD format
        '''
        today = datetime.datetime.utcnow()
        today_date = today.date()
        today = today_date.strftime('%Y%m%d')

        return str(today)

    def __insert_data_to_changelog(self, dataset_id, table_name,
                                   changelog_data):
        '''
            Function that inserts data into the changelog table
            Args:
                dataset_id        (string): The id of the data set
                table_name        (string): The name of the table
                changelog_data    (dict): The Patients Data
            Returns
                Boolean: True if the request worked, false if other wise
        '''
        parent                          = self.__google_big_query_storage_client\
            .table_path("indx-data-services", dataset_id, table_name)
        write_stream = types.WriteStream()
        write_stream.type_ = types.WriteStream.Type.COMMITTED
        write_stream                    = self.__google_big_query_storage_client\
            .create_write_stream(
            parent=parent, write_stream=write_stream
        )
        stream_name = write_stream.name

        # Create a template with fields needed for the first request.
        request_template = types.AppendRowsRequest()

        # The initial request must contain the stream name.
        request_template.write_stream = stream_name

        # So that BigQuery knows how to parse the serialized_rows, generate a
        # protocol buffer representation of your message descriptor.
        proto_schema = types.ProtoSchema()
        proto_descriptor = descriptor_pb2.DescriptorProto()
        changelog_schema_pb2.changelog_schema_pb2.DESCRIPTOR.CopyToProto(
            proto_descriptor)
        proto_schema.proto_descriptor = proto_descriptor
        proto_data = types.AppendRowsRequest.ProtoData()
        proto_data.writer_schema = proto_schema
        request_template.proto_rows = proto_data

        # Create a batch of row data by appending proto2 serialized bytes to the
        # serialized_rows repeated field.
        serialized_row = self.__create_proto_buf_row(changelog_data)
        response_future_1 = self.__big_query_wrapper.insert_to_bq_table(
            self.__google_big_query_storage_client, serialized_row,
            request_template, stream_name, parent)

        if response_future_1.done():
            self.__changelog_insert_logger.log_text(
                'Inserted ' + str(changelog_data) + ' to changelog',
                severity='INFO')
            return True

        self.__changelog_insert_logger.log_text(
            'Failed to insert ' + str(changelog_data) + ' to CHANGELOG table',
            severity='ERROR')
        return False

    def __create_proto_buf_row(self, protobuf_data):
        '''
            Creates protobuf object to be inserted into the data lake changelog table
            Args:
                protobuf_data (dict): The data we wish to insert
            returns
                serialized proto buf: The serialized proto buf object of the passed
                    dictionary
        '''
        row = changelog_schema_pb2.changelog_schema_pb2()
        for field in self.table_fields:

            if field in protobuf_data:
                row = self.__proto_buf_dictionary[field](row,
                                                         protobuf_data[field])

        return row.SerializeToString()

    def __insert_client_id(self, row, client_id):
        if client_id is None:
            return row
        row.CLIENT_ID = client_id
        return row

    def __insert_bucket_id(self, row, bucket_id):
        if bucket_id is None:
            return row
        row.BUCKET_ID = bucket_id
        return row

    def __insert_file_path(self, row, file_path):
        if file_path is None:
            return row
        row.FILE_PATH = file_path
        return row

    def __insert_file_name(self, row, file_name):
        if file_name is None:
            return row
        row.FILE_NAME = file_name
        return row

    def __insert_action(self, row, action):
        if action is None:
            return row
        row.ACTION = action
        return row

    def __insert_action_at(self, row, action_at):
        if action_at is None:
            return row
        row.ACTION_AT = action_at
        return row

    def __insert_replicated(self, row, replicated):
        if replicated is None:
            return row
        row.REPLICATED = replicated
        return row

    def __insert_replicated_at(self, row, replicated_at):
        if replicated_at is None:
            return row
        row.REPLICATED_AT = replicated_at
        return row



#
#   Created By Ryan McDermott
#   Created on 2/18/2022
#

import csv
import os
from pathlib import Path
import unicodedata

from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClientWrappers import CloudStorage
from indxdatalaketools.ArgumentValidation import PatientData
from indxdatalaketools.ArgumentValidation import ClientId
from indxdatalaketools.ArgumentValidation import FileMetadata
from indxdatalaketools.ArgumentValidation import ValidationHelpers
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients
from indxdatalaketools.PropertiesDBApi import Properties
from pyrsistent import b

DOES_NOT_EXIST_BAD_REQUEST = " does not exist: Bad Request"


class Validator():
    '''
        Command line tool helper class to verify all command line arguments are
        valid
    '''
    arguments = {}
    data_lake_tool_command = ''

    __argument_validation_functions = None

    def __init__(self):
        '''
            Class that validates all arguments passed.
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
            Returns:
                None
        '''

        self.__set_up_argument_validation_functions()

    def __set_up_argument_validation_functions(self):
        '''
            Sets up the dictionary containting all valid arguments and the 
            functions needed to validate them
        '''
        self.__argument_validation_functions = {
            'client_id': self.__validate_client_id,
            'patient_data': self.__validate_patient_data,
            'modality': self.__validate_modality,
            'mrn': self.__validate_mrn,
            'file_metadata': self.__validate_file_metadata,
            'file_path': self.__validate_file_path,
            'file_name': self.__validate_file_name,
            'gsutil_uri': self.__validate_gsutil_uri,
            'patient_id': self.__validate_patient_id,
            'restore_date': self.__validate_restore_date,
            'incremental_restore_date':
            self.__validate_incremental_restore_date,
            'archive_bucket': self.__validate_archive_bucket,
            'batch_file': self.__validate_batch_file
        }

    def validate_all_arguments(self, arguments, data_lake_tool_command):
        '''
            Verifys all arguments passed, some arguments such as CLIENT_ID have 
            different validation checks depending on the command called
            Args:
                arguments               (dict): key/value pair of all arguments passed
                data_lake_tool_command  (string): The command called by the data lake tools
            Returns:
                boolean: True if all the arguments are valid, False if otherwise
        '''
        self.arguments = arguments
        self.data_lake_tool_command = data_lake_tool_command

        for argument in arguments:
            if not self.__validate_argument(argument):
                return False

        return True

    def __validate_argument(self, argument):
        '''
            Validate a single argument based on they key passed
            Args:
                argument (string): The key of the argument for the arguments dictionary
            Returns:
                boolean: True if the argument is valid, False if otherwise
        '''
        return self.__argument_validation_functions[argument]()

    def __validate_client_id(self):
        '''
            Validates the client id argument, and exists in Datastore
            Args:
                None
            Returns:
                boolean: True if the client id is valid, False if otherwise
        '''
        return ClientId.Validator().validate(self.arguments['client_id'],
                                             self.data_lake_tool_command)

    def __validate_patient_data(self):
        '''
            Validates the patient data argument
            Args:
                None
            Returns:
                boolean: True if the patient data is valid, False if otherwise
        '''
        return PatientData.Validator().validate(self.arguments['patient_data'],
                                                self.arguments)

    def __validate_file_metadata(self):
        '''
            Validates the file metadata
            Args:
                None
            Returns:
                boolean: True if the modality is valid, False if otherwise
        '''
        return FileMetadata.Validator().validate(self.arguments,
                                                 self.data_lake_tool_command)

    def __validate_modality(self):
        '''
            Validates the modality
            Args:
                None
            Returns:
                boolean: True if the modality is valid, False if otherwise
        '''
        if self.arguments['modality'] in Properties.Properties.instance(
        ).modalities:
            return True

        Helpers.print_error('Incorrect modality ' + self.arguments['modality'])
        return False

    def __validate_mrn(self):
        '''
            Validates that the MRN only contains digits and letters
            Args:
                None
            Returns:
                boolean: True if the MRN is valid, False if otherwise
        '''
        if all(character.isalpha() or character.isdigit() or \
            character == '-' for character in self.arguments['mrn']):
            return True

        Helpers.print_error(self.arguments['mrn'] + \
            " is not a valid MRN: Does Not contain only letters, digits and hyphens")
        return False

    def __validate_patient_id(self):
        '''
            Validates the the patient id exists in the datalake
            Args:
                None
            Returns:
                boolean: True if the MRN is valid, False if otherwise
        '''
        patient_id = self.arguments['patient_id']
        client_id = self.arguments['client_id']
        for modality in Properties.Properties.instance().modalities:
            bucket = GcpClients.instance().get_storage_client().get_bucket(
                client_id)
            blob_name = modality + '/' + patient_id

            blob_list = list(bucket.list_blobs(prefix=blob_name))
            if len(blob_list) > 0:
                return True

        print(patient_id + ' is not a valid patient id')
        return False

    def __validate_file_name(self):
        '''
            Validates the file name of a gcs_blob
            Args:
                None
            Returns:
                boolean: True if the file exists, False if otherwise
        '''
        patient_id = Helpers.patient_hash(self.arguments['client_id'],
                                          [self.arguments['mrn']])[0]
        bucket_name = self.arguments['client_id']
        source_blob_name    = self.arguments['modality'] + \
            '/' + patient_id + '/' + self.arguments['file_name']
        bucket = GcpClients.instance().get_storage_client().bucket(bucket_name)
        blob = bucket.blob(source_blob_name)

        result = Helpers.try_blob_exists(blob)
        if result:
            return True

        Helpers.print_error(source_blob_name + DOES_NOT_EXIST_BAD_REQUEST)
        return False

    def __validate_file_path(self):
        '''
            Validate the File Path
            Args:
                None
            Returns:
                boolean: True if the file path is valid False if other wise
        '''
        # gcs bucket
        if 'gs://' in self.arguments['file_path']:
            return self.__if_blob_exists(self.arguments['file_path'])

        # local file
        local_file_path = Path(self.arguments['file_path'])
        result = local_file_path.is_file()

        if result:
            return True

        Helpers.print_error(self.arguments['file_path'] +
                            DOES_NOT_EXIST_BAD_REQUEST)
        return False

    def __validate_gsutil_uri(self):
        '''
            Validates the gsutil uri
            Args:
                None
            Returns:
                boolean: True if the gsutil rui is valid
        '''
        return self.__if_blob_exists(self.arguments['gsutil_uri'])

    def __if_blob_exists(self, gcs_gsutil_uri):
        '''
            Checks if a gcs blob exists
            Args:
                gcs_gsutil_uri (string) gsutil uri of blob we want to upload
            Returns:    
                boolean: True if the blob exists, False if otherwise
        '''
        file_path = gcs_gsutil_uri.replace("gs://", "").replace("\"", "")
        bucket_name = file_path.split("/")[0]
        source_blob_name = file_path.split("/", 1)[1]
        bucket = GcpClients.instance().get_storage_client().bucket(bucket_name)
        blob = bucket.blob(source_blob_name)

        result = Helpers.try_blob_exists(blob)
        if result:
            return True

        Helpers.print_error(source_blob_name + DOES_NOT_EXIST_BAD_REQUEST)
        return False

    def __validate_incremental_restore_date(self):
        '''
            Checks if a the provided restoration date is in the correct format and files exist for that date in the archive
            Args:
                None
            Returns:    
                boolean: True if the restoration date is valid, False if otherwise
        '''
        if not ValidationHelpers.validate_date(
                self.arguments['incremental_restore_date']):
            return False

        blob_list = GcpClients.instance().get_storage_client().list_blobs(
            'indx-dummy-archive',
            prefix=self.arguments['client_id'] + '/' +
            self.arguments['incremental_restore_date'] + '/')
        if len(list(blob_list)) == 0:
            return False

        return True

    def __validate_restore_date(self):
        ''' 
            Validates the restore date argument 
            returns:
                boolean: True if the restore date is valid, false if otherwise
        '''
        if not ValidationHelpers.validate_date(self.arguments['restore_date']):
            return False

        file_path = 'PATIENTS_BACKUP/UNKNOWN'
        if len(CloudStorage.Wrapper().find_blobs_with_metadata(
                self.arguments['client_id'],
                'SNAPSHOT_DATE',
                self.arguments['restore_date'],
                path_prefix=file_path)) > 0:

            return True

        return False

    def __validate_archive_bucket(self):
        ''' 
            Validates the restore date argument 
            returns:
                boolean: True if the archive bucket exists, false if otherwise
        '''
        bucket_name = self.arguments['archive_bucket']
        if not CloudStorage.Wrapper().\
            check_if_bucket_exists(bucket_name):
            return False

        if bucket_name != 'indx-datalake-archives' and bucket_name != 'indx-dummy-archive':
            Helpers.print_error(
                'archive bucket is not either indx-datalake-archives or indx-dummy-archive'
            )
            return False

        return True

    def __validate_batch_file(self) -> bool:
        ''' validates the batch file
    
        Returns:    
            bool: True if the batch file is valid, False if otherwise
        '''
        batch_file = self.arguments['batch_file']

        if not os.path.isfile(batch_file):
            Helpers.print_error("Cannot find provided batch file " + batch_file)
            return False

        file_list = Helpers.read_batch_file(batch_file)
        header = file_list[0]

        # check header
        if 'MODALITY' not in header or 'MRN' not in header \
            or 'METADATA' not in header or 'FILE_PATH' not in header:
            Helpers.print_error("Header in csv batch file is invalid")
            return False

        modality_index = header.index('MODALITY')
        mrn_index = header.index('MRN')
        metadata_index = header.index('METADATA')
        file_path_index = header.index('FILE_PATH')

        # validate infromation
        for i in range(1, len(file_list)):
            arguments = {
                'client_id': self.arguments['client_id'],
                'modality': file_list[i][modality_index],
                'mrn': file_list[i][mrn_index],
                'file_metadata': file_list[i][metadata_index],
                'file_path': file_list[i][file_path_index]
            }
            if not self.validate_all_arguments(arguments, 'file'):
                Helpers.print_error('Problem on line ' + str(i + 1))
                return False

        # no problems
        return True



#
#   Created by Ryan McDermott
#   Created on 5/12/2022
#

import datetime
import time
from indxdatalaketools import DataLakeGlobals
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients
from indxdatalaketools.PropertiesDBApi import Properties
from indxdatalaketools.ArgumentValidation import Argument

from indxdatalaketools.DataStandardizer import FileMetadata, PatientTable
from indxdatalaketools.ClientTools.Files import Upload
from indxdatalaketools.ClientTools.PatientsTable import Insert
from indxdatalaketools.DocumentClassifierApi import DocumentClassifierApi
from indxdatalaketools.GoogleClientWrappers import CloudStorage
from indxdatalaketools import Helpers
import warnings


class Client:

    client_uuid = ''

    __argument_validator = None

    def __init__(self, client_uuid, credentials_file_path=None):
        '''
            Init function that sets up client credentials, uuid, 
            and File operations API
            Args:
                credentials_file_path   (string): The path to a Service Account key json file
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid

        GcpClients.instance(credential_file=credentials_file_path)
        self.__client_ops_logger = GcpClients.instance().get_logging_client(
        ).logger(client_uuid + '-file-logger')
        if self.client_uuid == 'indx-test-client-datalake':
            self.__throughput_logger = GcpClients.instance().get_logging_client(
            ).logger(client_uuid + '-throughput')
        self.__argument_validator = Argument.Validator()
        self.__document_classifier = DocumentClassifierApi.DocClassifierClient()

    def upload_file(self, modality, mrn, metadata, file_path):
        '''
            Top level function that uploads a file to a Datalake, This will aslo add the 
            Patient data to the PATIENTS table if applicable
            Args:
                modality    (string): The modality of the file
                mrn         (string): The mrn of who the file belongs to
                metadata    (dict|string|file_path): Any metadata belonging to the file
                file        (string): The file path of the file we wish to upload
            Returns:
                boolean: True if the operation was a success, False if otherwise
        '''
        # validate arguments
        arguments = {
            'client_id': self.client_uuid,
            'modality': modality,
            'mrn': mrn,
            'file_metadata': metadata,
            'file_path': file_path
        }
        data_lake_command = 'file'

        Properties.Properties.instance()

        if not self.__argument_validator.validate_all_arguments(
                arguments, data_lake_command):
            return False

        return self.__standardize_and_run_file_upload(arguments)

    def batch_file_upload(self, file_path: str) -> bool:
        ''' Uploads multiple files found in a csv file

        Top level function that uploads multiple files to a datalake in one command,
        The file must contain the modality, mrn, metadata and file path for each file.

        Args:
            - file_path (str): The path to the batch file
        
        Returns:
            bool: True if all files were uploaded to the datalake, False if otherwise
        '''
        arguments = {'client_id': self.client_uuid, 'batch_file': file_path}
        data_lake_command = 'file'

        start = time.time()
        Properties.Properties.instance()

        if not self.__argument_validator.validate_all_arguments(
                arguments, data_lake_command):
            return False

        list_of_files = Helpers.read_batch_file(file_path)

        if self.client_uuid == 'indx-test-client-datalake':
            finish = time.time()
            upload_time = finish - start
            self.__throughput_logger.log_text(
                "It took " + str(upload_time) + " seconds to validate " +
                str(len(list_of_files) - 1) + " records",
                severity="INFO")

        failed_uploads = self.__upload_files_from_list(list_of_files, start)

        if len(failed_uploads) == 0:
            return True

        for i in failed_uploads:
            Helpers.print_error("Failure uploading patient " + i[1]['mrn'] +
                                " on line " + str(i[0]))

        return False

    def __upload_files_from_list(self, list_of_files, start):
        '''loops through passed list and inserts files to the datalake

        This function loops through the list passed and insertes each file into the clients datalake
        all errors and upload times are logged.
        
        Args:
            list_of_files (list): the list of file data we wish to upload
            start (time.time): the start time of the operation
        Returns:
            list: the list of files that did not get processed

        '''
        header = list_of_files[0]
        failed_uploads = []
        modality_index = header.index('MODALITY')
        mrn_index = header.index('MRN')
        metadata_index = header.index('METADATA')
        file_path_index = header.index('FILE_PATH')

        # skip header
        for row in range(1, len(list_of_files)):
            modality = list_of_files[row][modality_index]
            mrn = list_of_files[row][mrn_index]
            metadata = list_of_files[row][metadata_index]
            file_path = list_of_files[row][file_path_index]

            arguments = {
                'client_id': self.client_uuid,
                'modality': modality,
                'mrn': mrn,
                'file_metadata': metadata,
                'file_path': file_path
            }
            result = self.__standardize_and_run_file_upload(arguments)

            if not result:
                failed_uploads.append((row, arguments))

            if result and self.client_uuid == 'indx-test-client-datalake':
                finish = time.time()
                upload_time = finish - start
                self.__throughput_logger.log_text(
                    str(row) + " (" + str(DataLakeGlobals.SIZE_OF_FILES) +
                    " MB) files have been processed in " + str(upload_time) +
                    "seconds",
                    severity="INFO")

        return failed_uploads

    def upload_patient(self, patient_data):
        '''
            Top level function that uploads a patient to a PATIENT table, This will also update missing
            patient data but not overwrite it.
            Args:
                patient_data (dict): Dictionary containing the data to insert a patient to the patient table
            Returns:
                boolean: True if the operation was a success, False if otherwise
        '''
        # verify inputs
        arguments = {
            'client_id': self.client_uuid,
            'patient_data': patient_data
        }
        data_lake_command = 'patients-table'

        Properties.Properties.instance()

        if not self.__argument_validator.validate_all_arguments(
                arguments, data_lake_command):
            return False

        # clean data
        data_standardizer = PatientTable.Client()
        patient_data = data_standardizer.standardize(patient_data)

        # run the command
        return Insert.Client(
            self.client_uuid).insert_data_to_patients_table(patient_data)

    def __document_classifier_operations(self, mrn, file_path):
        '''
            function that handles operations required to run the document classifier. This includes uploading file from local to 
            a staging folder on the clients data lake, running the document classifier with this uploaded blob and then deleting the blob
            args:
                mrn: str - input mrn
                file_path: str - local file path of input file
            return:
                classified_modality: str - modaity prediction of the document classifier
        '''
        blob = CloudStorage.Wrapper().upload_file_to_gcs_bucket(
            mrn, file_path, "staging_folder", self.client_uuid)
        if not blob:
            return blob

        gcs_file_path = "gs://" + blob.bucket.name + "/" + blob.name
        classified_modality, classified_file_type = self.__send_request_to_document_classifier(
            gcs_file_path)

        return classified_modality

    def __standardize_and_run_file_upload(self, arguments: dict) -> bool:
        ''' Standardizes the arguments and data then uploads the file to the datalake.

        standradizes the agruments passed then calls the file upload command.

        Args:
            arguments (dict): the arguments to be passed to the file upload command

        Returns:
            bool: True if the file was successfully uploaded. False if otherwise
        '''
        # standardize the metadata
        file_metadata_standardizer = FileMetadata.Standardizer(arguments)
        file_metadata = file_metadata_standardizer.standardize(
            arguments['file_metadata'])

        # run the document classifier, check if modality matches predicted modality
        classified_modality = self.__document_classifier_operations(
            arguments['mrn'], arguments['file_path'])
        if not classified_modality or classified_modality is None:
            classified_modality = 'Failed To Classify'
            warnings.warn("Request to document classifier failed.")
            self.__client_ops_logger.log_text(
                'failed to run document classifier on file ' +
                arguments['file_path'],
                severity='WARNING')
        elif classified_modality != arguments['modality']:
            Helpers.print_error(
                "Modality may be incorrect. Modality might be " +
                classified_modality)

        file_metadata['CLASSIFIED_MODALITY'] = classified_modality

        # run command
        return Upload.Client(self.client_uuid).upload_file_to_datalake(
            arguments['modality'], arguments['mrn'], file_metadata,
            arguments['file_path'])

    def __send_request_to_document_classifier(self, gcs_file_path):
        '''
            runs the document classifier
            args:
                gcs_file_path: str gcs path of file in staging folder
            return:
                classified_modality: str - modality prediction of the document classifier
                classified_file_type: str - file type prediction of the document classifier
        '''
        try:
            return self.__document_classifier.get_modality_and_file_type(
                gcs_file_path)
        except Exception as e:
            print("Error when trying to run document classifier. " + str(e))
            return False, False


#
#   Created By Ryan McDermott
#   Created on 2/21/2022
#

import datetime
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients
from indxdatalaketools.Helpers import print_error
from indxdatalaketools.GoogleClientWrappers import BigQuery
from indxdatalaketools.PatientsHoldingApi import HoldingTable


class Client:
    '''
        Class that handles all functionlaity to update a patient in the PATIENTS table
    '''
    client_uuid = ''
    dataset_id = ''
    patient_table_name = ''
    overwrite = False

    __patient_insert_logger = None
    __big_query_wrapper = None

    def __init__(self, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google big query client used to insert items to the Patients table
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid
        self.dataset_id = "indx-data-services."+self.client_uuid.\
            replace("-", "_")
        self.patient_table_name = self.dataset_id + ".PATIENTS"
        logging_client = GcpClients.instance().get_logging_client()
        self.__patient_insert_logger = logging_client.logger(client_uuid +
                                                             '-patient-logger')

    def update_patient_in_patient_table(self,
                                        patient_data,
                                        patient_row,
                                        overwrite=False):
        '''
            function that checks to see if the patient in the table needs to get updated or not
            and then applies the update if applicable.
            Args:
                patient_data    (dict): The information for the patient
                patient_row     (row): The row that we wish to update in the Table
                overwrite       (boolean): Flag to indicate whether or not we can over 
                    write the patients data
            Returns:
                boolean: True if the update was successful or if the patient 
                    did not need to be updated, False if otherwise
        '''
        self.overwrite = overwrite
        if self.__compare_dict_to_table_row(patient_data, patient_row):
            # The Patient Data is updatable so update
            time_now = datetime.datetime.utcnow()
            patient_update_time = datetime.datetime.fromisoformat(
                str(patient_row.get("UPDATED_AT")))
            if time_now < patient_update_time + datetime.timedelta(minutes=90):

                patient_data['UPDATED_AT'] = None
                patient_data = Helpers.convert_keys_in_dictionary(patient_data)
                patient_data['action'] = "UPDATED"
                self.__patient_insert_logger.log_text(
                    'staging patient data ' + str(patient_data) + " for update",
                    severity='INFO')
                return HoldingTable.Client().insert_data_to_holding_table(
                    patient_data)

            return self.__update_patients_table_row_with_dml(
                patient_data, patient_row=patient_row)

        print("Nothing to Update for Patient " + patient_data["PATIENT_ID"])
        return True

    def __update_patients_table_row_with_dml(self,
                                             patient_data,
                                             patient_row=None):
        '''
            This function updates the row of a table with new data. CLIENT_ID, PATIENT_ID,
            and MRN cannot be updated. This function builds the DPL query based on what the client
            wishes to update. Only values that are null in the PATIENTS table will be updated. if 
            overwrite is False then patient row must be passed.
            Args:
                patient_data    (dict): The new data that needs to be updated
                patient_row     (row): The row found in the big query table
            Returns:
                Boolean: True if the operation was succesfull, false if otherwise
        '''
        dml_query = self.__create_dml_update_query(patient_data, patient_row)
        if BigQuery.Wrapper().perform_query(dml_query) is None:
            self.__patient_insert_logger.log_text(
                'could not update patient info ' + str(patient_data),
                severity='WARNING')
            return False

        self.__patient_insert_logger.log_text('updated patient info ' +
                                              str(patient_data),
                                              severity='INFO')
        return True

    def __create_dml_update_query(self, patient_data, patient_row=None):
        '''
            Creates the dlp query necessary to update a Patient in the PATIENTS table
            Args:
                patient_data    (dict): The new data that needs to be updated
                patient_row     (row): The row found in the big query table
                overwrite       (boolean): Flag to indicate whether or not we can over 
                    write the patients data
            Returns:
                string: The query we wish to run on the PATIENTS table
        '''
        table_name = "indx-data-services." + self.client_uuid.replace(
            '-', '_') + ".PATIENTS"
        dml_query = 'UPDATE `' + table_name + '` SET '

        for k in patient_data:
            dml_query += self.__populate_changed_keys_in_query(
                k, patient_row.get(k), patient_data[k])

        # remove last comma
        dml_query = dml_query[:-2] + ' '
        dml_query += 'WHERE PATIENT_ID = \'' + patient_data['PATIENT_ID'] + '\''
        return dml_query

    def __populate_changed_keys_in_query(self, key, patient_row_value,
                                         patient_data_value):
        '''
            Populates all keys that are to be updated for the patient
            Args:
                key                (string): The key we are using to compare the Patients
                    Data
                patient_row_value  (string): The value found in the patient row for a 
                    given key
                patient_data_value (stirng): The value passed in the json for a given
                    key
            Returns:
                string: The new string to append to the dlp query, empty if nothing needs 
                    to be added
        '''
        #Client ID, Patient ID, and MRN are immutable
        if key == "CLIENT_ID" or key == "MRN" or key == "PATIENT_ID":
            return ''

        if self.overwrite == False and patient_data_value is None:
            return ''

        # New information can always be added to a patient record
        if self.overwrite == False and patient_row_value is None:
            return key + ' = \'' + str(patient_data_value) + '\', '

        # Overwriting existing information requires a special command
        if self.overwrite == True and patient_data_value is not None:
            return key + ' = \'' + str(patient_data_value) + '\', '

        # overwriting existing
        if self.overwrite == True and patient_data_value is None:
            return key + ' = DEFAULT, '

        return ''

    def __compare_dict_to_table_row(self, patient_data, patient_row):
        '''
            Function that compares the client provided json data to the information
            contained in the PATIENTS table for a specific PATIENT_ID. If
            they are the same then we do not need to update. If the PATIENTS table
            contains more information we do not need to update.
            Args:
                patient_data    (dict): The dictionary containing the client provided patient
                    informations
                patient_row     (row): The information found in the PATIENTS table for a specific
                    PATIENT_ID
            Returns:
                True if a patient needs to be updated, false if otherwise.
        '''

        status = True

        if self.overwrite == True:
            return True

        if self.__compare_checksum(patient_data, patient_row):
            print("No Changes in Checksum")
            return False

        # compare, the changes in the two records
        for k in patient_data:
            if k == "CLIENT_ID" or k == "MRN" or k == "PATIENT_ID" \
            or k == "UPDATED_AT" or k == "CHECKSUM":
                continue

            elif not self.__is_updateable(patient_data[k], patient_row.get(k)):
                status = False

        return status

    def __compare_checksum(self, patient_data, patient_row):
        '''
            Function that compares the tnew value and old values checksum to see if 
            the values are the same or not.
            Args:
                patient_data    (dict): The dictionary containing the client provided patient
                    informations
                patient_row     (row): The information found in the PATIENTS table for a specific
                    PATIENT_ID
            Returns:
                boolean: True if the values are the same, False if they are different
        '''
        if patient_data['CHECKSUM'] == patient_row.get('CHECKSUM'):
            return True

        return False

    def __is_updateable(self, input_value, existing_value):
        '''
            Function that checks whether or not there is new data in the
            json passed for a patient. If the the values are the same True is returned. 
            Args:
                patient_json_value (string): The value of the patient json data passed in
                existing_value  (string): The value already existing in the PATIENTS table
            Returns:
                boolean: True if there is more data passed, False if not
        '''
        if input_value is None and existing_value is None:
            # no values, is updateable
            return True

        if input_value is None and existing_value is not None and self.overwrite == False:
            # no trying to replace value this none, not updateable
            return False

        if input_value is not None and existing_value is None:
            # can update the patient
            return True

        if input_value is not None and existing_value is not None and self.overwrite == False:
            if input_value == existing_value:
                # no change here
                return True
            print(input_value, existing_value)
            # different values cannot overwrite, not updateable
            print_error(
                "Cannot overide values in the PATIENT table using this command,"
                + " try patients-table update command instead.")
            return False

        # default changeable
        return True


#
#   Created by Ryan McDermott
#   Created on 6/13/2022
#

import json
from google.cloud import tasks_v2
from indxdatalaketools.GoogleClientWrappers import CloudTasks
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients

HOLDING_TABLE_URL = 'https://patients-db-api-6vyiigtwia-uc.a.run.app'


class Client:
    '''
        Class that contains all the functionality make a call to the 
        internal Cloud Run API service
    '''
    __cloud_tasks_client = None
    __cloud_tasks_wrapper = None

    def __init__(self):
        '''
            Init function that sets up the Holding Table object

            Returns:
                None
        '''
        self.__cloud_tasks_client = GcpClients.instance().get_tasks_v2_client()
        self.__cloud_tasks_wrapper = CloudTasks.Wrapper()

    def create_holding_table_if_not_exist(self, table_data):
        '''
            Function that creates a cloud sql table if it does not already exist
            Args:
                - table_data (dict): The values of the table to create
            Returns:
                - boolean: True if the table got created false if otherwise
        '''
        payload = json.dumps(table_data)
        parent  = self.__cloud_tasks_client.\
            queue_path("indx-data-services", "us-central1", "PATIENTS-holding-table-flush")
        task = {
            "http_request": {  # Specify the type of request.
                "http_method": tasks_v2.HttpMethod.POST,
                "url":
                'https://patients-db-api-6vyiigtwia-uc.a.run.app/patientTable',  # The full url path that the task will be sent to.
                "oidc_token": {
                    "service_account_email":
                    'patients-holding-table@indx-data-services.iam.gserviceaccount.com',
                    "audience":
                    'https://patients-db-api-6vyiigtwia-uc.a.run.app'
                }
            }
        }

        task["http_request"]["headers"] = {"Content-type": "application/json"}
        converted_payload = payload.encode()
        task["http_request"]["body"] = converted_payload

        return self.__cloud_tasks_wrapper.create_http_task(parent, task)

    def delete_holding_table(self, client_id):
        '''
            Function that deletes a cloud sql table if it does not already exist
            Args:
                - table_data (dict): The values of the table to create
            Returns:
                - boolean: True if the table got created false if otherwise
        '''
        parent  = self.__cloud_tasks_client.\
            queue_path("indx-data-services", "us-central1", "PATIENTS-holding-table-flush")
        task = {
            "http_request": {  # Specify the type of request.
                "http_method":
                tasks_v2.HttpMethod.DELETE,
                "url":
                'https://patients-db-api-6vyiigtwia-uc.a.run.app/patientTable/'
                + client_id,  # The full url path that the task will be sent to.
                "oidc_token": {
                    "service_account_email":
                    'patients-holding-table@indx-data-services.iam.gserviceaccount.com',
                    "audience":
                    'https://patients-db-api-6vyiigtwia-uc.a.run.app'
                }
            }
        }

        return self.__cloud_tasks_wrapper.create_http_task(parent, task)

    def insert_data_to_holding_table(self, patient_data):
        '''
            Function that inserts patient data to the holding table. If
                the patient record already exists, it is updated.
            Args:
                - patient_data (dict): The values of the patient to create
            Returns:
                - boolean: True if the patient data was created, False if otherwise
        '''
        client_id = patient_data['client-id']
        payload = json.dumps(patient_data)
        parent  = self.__cloud_tasks_client.\
            queue_path("indx-data-services", "us-central1", "PATIENTS-holding-table-flush")
        task = {
            "http_request": {  # Specify the type of request.
                "http_method":
                tasks_v2.HttpMethod.POST,
                "url":
                'https://patients-db-api-6vyiigtwia-uc.a.run.app/patientEntry/'
                + client_id,  # The full url path that the task will be sent to.
                "oidc_token": {
                    "service_account_email":
                    'patients-holding-table@indx-data-services.iam.gserviceaccount.com',
                    "audience":
                    'https://patients-db-api-6vyiigtwia-uc.a.run.app'
                }
            }
        }

        task["http_request"]["headers"] = {"Content-type": "application/json"}
        converted_payload = payload.encode()
        task["http_request"]["body"] = converted_payload

        return self.__cloud_tasks_wrapper.create_http_task(parent, task)



#
#   Created by Arjun Acharya
#   Created on 7/27/2022
#

from indxdatalaketools.GoogleClientWrappers import SecretManager
from indxdatalaketools import Helpers
import json

DOCUMENT_CLASSIFIER_URL = 'https://documentclassifier.indx.acuityeyegroup.dev/'


class DocClassifierClient:
    '''
    Class that contains all the functionality make a call to Document Classifier
    '''
    __project_id = ''
    __secret_id = ''

    def __init__(self):
        self.__project_id = 'indx-data-services'
        self.__secret_id = 'indx-document-classifier-api-key'

    def get_modality_and_file_type(self, gcs_file_path):
        '''
            creates the request to the document classifier cloud run service
            args:
                gcs_file_path: str - gcs path of file on the staging bucket
            return:
                classified_modality: str - modality prediction made by the document classifier
                classified_file_type: str - file type prediction made by the document classifier
        '''
        api_key = SecretManager.Wrapper().get_secret(self.__project_id,
                                                     self.__secret_id)

        headers = {"Content-Type": "application/json", "x-api-key": api_key}

        payload = {"gcs_file_path": gcs_file_path}
        data = json.dumps(payload)
        data = data.encode('utf-8')
        response = Helpers.create_post_request(DOCUMENT_CLASSIFIER_URL +
                                               "predict",
                                               headers=headers,
                                               body=data)
        if response is None:
            return None, None
        classified_modality = response['modality']
        classified_file_type = response['file_type']

        return classified_modality, classified_file_type


#
#   Created by Ryan McDermott
#   Created on 7/7/2022
#

import google_crc32c
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients
from indxdatalaketools.Helpers import print_error


class Wrapper:
    ''' Wrapper for all secret manager client API Calls'''

    def __init__(self):
        ''' Nothin to initialize'''
        pass

    def get_secret(self, project_id, secret_id, version_id='latest'):
        '''
            Function that returns a secret found in secret manager
            Args:
                project_id (string): The project id where the secret is
                secret_id (string): The id of the secret
                version_id (string): The version of the sercet, defauls to latest
            Returns:
                string: The secret
        '''
        try:
            name = f"projects/{project_id}/secrets/{secret_id}/versions/{version_id}"
            response = GcpClients.instance().get_secretmanager_client().\
                access_secret_version(request={"name": name})

            # Verify payload checksum.
            crc32c = google_crc32c.Checksum()
            crc32c.update(response.payload.data)
            if response.payload.data_crc32c != int(crc32c.hexdigest(), 16):
                print_error("Data corruption detected for secret id" +
                            secret_id)
                return ''

            payload = response.payload.data.decode("UTF-8")
            return payload
        except Exception as err:
            print_error('problem accessing secret ' + name + ' ' + str(err))
            return ''



#
#   Created by Ryan McDermott
#   Created on 2/21/2022
#

import datetime
import sys
from google.cloud.bigquery_storage_v1 import types
from google.protobuf import descriptor_pb2

from indxdatalaketools.ClientTools.PatientsTable import table_schema_pb2
from indxdatalaketools.ClientTools.PatientsTable import Update
from indxdatalaketools import Helpers
from indxdatalaketools.GoogleClientWrappers import BigQuery
from indxdatalaketools.GoogleClients.GoogleClients import GcpClients


class Client:
    '''
        Class that contains all the functionality to insert a patient into
        The patient table
    '''

    client_uuid = ''
    dataset_id = ''
    patient_table_name = ''
    table_fields = [
        "CLIENT_ID", "MRN", "PATIENT_ID", "FIRSTNAME", "MIDDLENAME", "LASTNAME",
        "DOB", "SEX", "RACE", 'CHECKSUM', 'UPDATED_AT'
    ]
    __patient_insert_logger = None
    __proto_buf_dictionary = None
    __google_big_query_client = None
    __google_big_query_storage_client = None
    __big_query_wrapper = None
    __patient_table_update_client = None

    def __init__(self, client_uuid):
        '''
            Init function that sets up client credentials, uuid, 
            and google api clients used to insert items to the Patients table
            Args:
                google_clients          (ApiClients): object containing all Google Api Clients
                client_uuid             (string): The client's uuid
            Returns:
                None
        '''
        self.client_uuid = client_uuid
        self.dataset_id = self.client_uuid.replace("-", "_")
        self.patient_table_name                 = 'indx-data-services.' + \
            self.dataset_id + ".PATIENTS"

        logging_client = GcpClients.instance().get_logging_client()
        self.__patient_insert_logger = logging_client.logger(client_uuid +
                                                             '-patient-logger')
        self.__create_proto_buf_dictionary()

    def __create_proto_buf_dictionary(self):
        '''
            Function that creates the protobuf ditionary to use to populate 
            its attributes easily
        '''
        self.__proto_buf_dictionary = {
            "CLIENT_ID": self.__insert_client_id,
            "MRN": self.__insert_mrn,
            "PATIENT_ID": self.__insert_patient_id,
            "FIRSTNAME": self.__insert_first_name,
            "MIDDLENAME": self.__insert_middle_name,
            "LASTNAME": self.__insert_last_name,
            "DOB": self.__insert_dob,
            "SEX": self.__insert_sex,
            "RACE": self.__insert_race,
            "CHECKSUM": self.__insert_checksum,
            "UPDATED_AT": self.__insert_updated_at,
        }

    def insert_data_to_patients_table(self, patient_json, overwrite=False):
        '''
            This function determines whether the patient_json is a list
            or a single patient. Then it uploads all patient data to the PATIENT 
            table for the client_id
            Args:
                patient_json    (json): Either a json object or a JSON Array containing
                    multiple patients data
                overwrite       (boolean): Flag to indicat whether or not we can overwrite the patients 
                    data
            Returns: 
                boolean: True if the patient/s were succesfully inserted in the PATIENTS table
        '''
        # parse json to determine single or multiple patients

        # multiple patients
        if isinstance(patient_json, list):
            return self.__insert_multiple_patients_to_table(patient_json,
                                                            overwrite=overwrite)

        return self.__insert_patient_to_patients_table(patient_json,
                                                       overwrite=overwrite)

    def __insert_multiple_patients_to_table(self,
                                            json_structure,
                                            overwrite=False):
        '''
            loops through json structure and inserts all patients into the PATIENTS
            table, if a patient is not successfully uploaded then it is logged and
            the process continues.
            Args:
                json_structure (dict[]): list of dictionarys containing the patients data
                overwrite (boolean): Flag to indicate whether or not the row needs to get
                    overwritten
            Returns:
                boolean: True if the operation was a success False if otherwise
        '''
        status = True
        for i in json_structure:
            if self.__insert_patient_to_patients_table(i, overwrite=overwrite):
                continue
            else:
                status = False

        if status == False:
            self.__patient_insert_logger.log_text(
                'One or more patients have not been uploaded Successfully',
                severity='ERROR')

        self.__patient_insert_logger.log_text(
            'All patients have been uploaded Successfully', severity='INFO')
        return status

    def __insert_patient_to_patients_table(self, patient_data, overwrite=False):
        '''
            Function that adds a single patient to the patient table, if that patient already exists
            in the table their data is updated with the provided information
            Args:
                patient_data    (dict): The data we wish to add to the patient table
                overwrite       (boolean): Flag to indicate whether or not we can over 
                    write the patients data
            Returns:
                boolean: True if the operation was a success, false if otherwise
        '''
        now = datetime.datetime.utcnow()
        patient_data['UPDATED_AT'] = str(now).replace(' ', 'T')
        patient_mrn = patient_data['MRN']
        patient_id = Helpers.patient_hash(self.client_uuid, [patient_mrn])[0]
        patient_row = self.__get_patient_in_patients_table(
            patient_mrn, patient_id)

        return self.__insert_data_to_patients_table(patient_data,
                                                    patient_row,
                                                    overwrite=overwrite)

    def __get_patient_in_patients_table(self, patient_mrn, patient_id):
        '''
            performs a query look up on the PATIENTS table to find if the given patient
            exists in the PATIENTS table
            Args:
                patient_mrn (string): The patients mrn
                patient_id  (string): The patients UUID
            Returns:
                The row of the Patient, if they are not in the table None is returned
        '''
        query = 'SELECT * \
                    FROM `' + self.patient_table_name + '`\
                    WHERE MRN =\'' + patient_mrn + '\'\
                    AND PATIENT_ID =\'' + patient_id + '\''
        rows = BigQuery.Wrapper().perform_query(query)
        if rows is None:
            return None

        if rows.total_rows == 0:
            return None
        if rows.total_rows > 1:
            Helpers.print_error(
                "CRITICAL: Error in PATIENTS table, multiple instances of MRN: "
                + patient_mrn + " and PATIENT_ID " + patient_id)

        # must only contain one row if not 0 and > 1
        for row in rows:
            return row

    def __insert_data_to_patients_table(self,
                                        patient_data,
                                        patient_row,
                                        overwrite=False):
        '''
            Inserts data to the patients table, if the data needs to be updated it will overwrite
            if the flag is set
            Args:
                patient_data (dict): The dictionary of the patients data
                patient_row  (row): The row already existing in the patients table
                overwrite    (boolean): True if we wish to overwrite the existing data in the 
                    PATIENTS table
            Returns:
                boolean: True if the operations was successful, False if otherwise
        '''
        if patient_row is not None:
            # patient exists, updated patient in table
            return Update.Client(
                self.client_uuid).update_patient_in_patient_table(
                    patient_data, patient_row, overwrite=overwrite)
        elif overwrite == False:
            # if not append to PATIENTS table
            return self.__add_data_to_patient_table(patient_data)
        else:
            Helpers.print_error("patient does not exist cannot update")
            return False

    def __add_data_to_patient_table(self, patient_data):
        '''
            Adds the patients data to the PATIENT table using the Big Query 
            Storage API COMMITTED row type to bypass the streaming buffer
            Args:
                patient_data    (dict): The Patients Data
            Returns
                Boolean: True if the request worked, false if other wise
        '''
        parent                          = GcpClients.instance().get_bigquery_storage_v1_client()\
            .table_path("indx-data-services", self.dataset_id, "PATIENTS")
        write_stream = types.WriteStream()
        write_stream.type_ = types.WriteStream.Type.COMMITTED
        write_stream                    = GcpClients.instance().get_bigquery_storage_v1_client()\
            .create_write_stream(
            parent=parent, write_stream=write_stream
        )
        stream_name = write_stream.name

        # Create a template with fields needed for the first request.
        request_template = types.AppendRowsRequest()

        # The initial request must contain the stream name.
        request_template.write_stream = stream_name

        # So that BigQuery knows how to parse the serialized_rows, generate a
        # protocol buffer representation of your message descriptor.
        proto_schema = types.ProtoSchema()
        proto_descriptor = descriptor_pb2.DescriptorProto()
        table_schema_pb2.table_schema_pb2.DESCRIPTOR.CopyToProto(
            proto_descriptor)
        proto_schema.proto_descriptor = proto_descriptor
        proto_data = types.AppendRowsRequest.ProtoData()
        proto_data.writer_schema = proto_schema
        request_template.proto_rows = proto_data

        # Create a batch of row data by appending proto2 serialized bytes to the
        # serialized_rows repeated field.

        serialized_row = self.__create_proto_buf_row(patient_data)
        if serialized_row is None:
            return False
        response_future_1 = BigQuery.Wrapper().insert_to_bq_table(
            GcpClients.instance().get_bigquery_storage_v1_client(),
            serialized_row, request_template, stream_name, parent)

        if response_future_1.done():
            self.__patient_insert_logger.log_text('successfully inserted ' +
                                                  str(patient_data) +
                                                  ' to PATIENTS table',
                                                  severity='INFO')
            print('successfully inserted Patient ' + str(patient_data['MRN']) +
                  ' to PATIENTS table')
            return True

        Helpers.print_error('Failed to insert Patient ' +
                            str(patient_data['MRN']) + ' to PATIENTS table')
        self.__patient_insert_logger.log_text(
            'Failed to insert ' + str(patient_data) + ' to PATIENTS table',
            severity='ERROR')
        return False

    def __create_proto_buf_row(self, patient_data):
        '''
            Creates a protobuf table_schema object to insert into a
            BQ table
            Args:
                patient_data (dict): Dictionary containing patient data
            Returns:
                serialized proto buf: The serialized proto buf object of the passed
                    dictionary
        '''
        row = table_schema_pb2.table_schema_pb2()
        for field in self.table_fields:

            if field in patient_data:
                row = self.__proto_buf_dictionary[field](row,
                                                         patient_data[field])

        try:
            return row.SerializeToString()
        except Exception as e:
            self.__patient_insert_logger.log_text(
                "error creating protocol buffer " + str(e), severity='ERROR')
            return None

    def __insert_client_id(self, row, client_id):
        if client_id is None:
            return row
        row.CLIENT_ID = client_id
        return row

    def __insert_mrn(self, row, mrn):
        if mrn is None:
            return row
        row.MRN = mrn
        return row

    def __insert_patient_id(self, row, patient_id):
        if patient_id is None:
            return row
        row.PATIENT_ID = patient_id
        return row

    def __insert_first_name(self, row, first_name):
        if first_name is None:
            return row
        row.FIRSTNAME = first_name
        return row

    def __insert_middle_name(self, row, middle_name):
        if middle_name is None:
            return row
        row.MIDDLENAME = middle_name
        return row

    def __insert_last_name(self, row, last_name):
        if last_name is None:
            return row
        row.LASTNAME = last_name
        return row

    def __insert_dob(self, row, dob):
        if dob is None:
            return row

        dob = dob.split("-")
        year = int(dob[0])
        month = int(dob[1])
        day = int(dob[2])
        date = datetime.date(year, month, day)
        past_date = datetime.date(1970, 1, 1)
        row.DOB = int((date - past_date).days)
        return row

    def __insert_sex(self, row, sex):
        if sex is None:
            return row
        row.SEX = sex
        return row

    def __insert_race(self, row, race):
        if race is None:
            return row
        row.RACE = race
        return row

    def __insert_checksum(self, row, checksum):
        if checksum is None:
            return row
        row.CHECKSUM = checksum
        return row

    def __insert_updated_at(self, row, updated_at):
        if updated_at is None:
            return row
        row.UPDATED_AT = updated_at
        return row


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: changelog_schema.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x63hangelog_schema.proto\"\xb0\x01\n\x14\x63hangelog_schema_pb2\x12\x11\n\tCLIENT_ID\x18\x01 \x02(\t\x12\x11\n\tBUCKET_ID\x18\x02 \x02(\t\x12\x11\n\tFILE_PATH\x18\x03 \x02(\t\x12\x11\n\tFILE_NAME\x18\x04 \x01(\t\x12\x0e\n\x06\x41\x43TION\x18\x05 \x01(\t\x12\x11\n\tACTION_AT\x18\x06 \x01(\t\x12\x12\n\nREPLICATED\x18\x07 \x01(\x08\x12\x15\n\rREPLICATED_AT\x18\x08 \x01(\t')



_CHANGELOG_SCHEMA_PB2 = DESCRIPTOR.message_types_by_name['changelog_schema_pb2']
changelog_schema_pb2 = _reflection.GeneratedProtocolMessageType('changelog_schema_pb2', (_message.Message,), {
  'DESCRIPTOR' : _CHANGELOG_SCHEMA_PB2,
  '__module__' : 'changelog_schema_pb2'
  # @@protoc_insertion_point(class_scope:changelog_schema_pb2)
  })
_sym_db.RegisterMessage(changelog_schema_pb2)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CHANGELOG_SCHEMA_PB2._serialized_start=27
  _CHANGELOG_SCHEMA_PB2._serialized_end=203
# @@protoc_insertion_point(module_scope)



import os
import setuptools
import site
import sys

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

dist = os.environ.get("INDX_DATALAKE_TOOLS_INSTALL", None)
print(dist)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

if dist == 'admin':
    setuptools.setup(
        name="indxdatalaketools",
        setuptools_git_versioning={
            "enabled": True,
        },
        license="Creative Commons Attribution-NonCommercial-NoDerivatives 4.0",
        license_files=('LICEN[CS]E*', 'COPYING*', 'NOTICE*', 'AUTHORS*'),
        setup_requires=["setuptools-git-versioning"],
        data_files=[
            ('man/man1', ['man/DataLakeTools.1']),
        ],
        author="Ryan McDermott",
        author_email="rmcdermott@intelligentdx.com",
        description=
        "Package that allows creation and manipulation of data lakes",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/IntelligentDX-LLC/data-lake-tools",
        project_urls={
            "Bug Tracker":
            "https://github.com/IntelligentDX-LLC/data-lake-tools/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        entry_points={
            'console_scripts': [
                'IndxAdminDataLakeTools = indxdatalaketools.AdminCommands.DataLakeToolsAdmin:main'
            ]
        },
        install_requires=[
            'google-cloud-datastore==2.5.1', 'google-cloud-storage==2.3.0',
            'google-cloud-bigquery==3.1.0',
            'google-cloud-bigquery-storage==2.13.1', 'google-auth==1.35.0',
            'google-cloud-pubsub==2.12.1', 'google-cloud-scheduler==2.6.3',
            'google-cloud-secret-manager==2.10.0',
            'google-cloud-logging==2.7.0', 'google-cloud-compute==0.7.0',
            'importlib-metadata==4.11.3', 'protobuf==3.20.1', 'click==8.1.3',
            'opencv-python==4.5.5.64', 'google-api-python-client==1.9.3',
            'google-cloud-tasks==2.7.1', 'requests==2.27.1', 'numpy==1.22.3',
            'google-crc32c==1.3.0'
        ],
        packages=setuptools.find_packages(exclude=[
            '**test**',
            '*GoogleCloudResources*',
        ]),
        python_requires=">=3.8",
        include_package_data=True,
        package_data={'': ['*ComputeEngineScripts/*']},
    )
else:
    setuptools.setup(
        name="indxdatalaketools",
        setuptools_git_versioning={
            "enabled": True,
            "template": "{tag}",
        },
        setup_requires=["setuptools-git-versioning"],
        license="Creative Commons Attribution-NonCommercial-NoDerivatives 4.0",
        license_files=('LICEN[CS]E*', 'COPYING*', 'NOTICE*', 'AUTHORS*'),
        data_files=[
            ('man/man1', ['man/DataLakeTools.1']),
        ],
        author="Ryan McDermott",
        author_email="rmcdermott@intelligentdx.com",
        description="Package that allows the upload of files to a datalake",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/IntelligentDX-LLC/data-lake-tools",
        project_urls={
            "Bug Tracker":
            "https://github.com/IntelligentDX-LLC/data-lake-tools/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
        ],
        entry_points={
            'console_scripts': [
                'IndxClientDataLakeTools = indxdatalaketools.ClientCommands.DataLakeToolsClient:main'
            ]
        },
        install_requires=[
            'google-cloud-datastore==2.5.1',
            'google-cloud-storage==2.3.0',
            'google-cloud-bigquery==3.1.0',
            'google-cloud-bigquery-storage==2.13.1',
            'google-auth==1.35.0',
            'google-cloud-pubsub==2.12.1',
            'google-cloud-scheduler==2.6.3',
            'google-cloud-secret-manager==2.10.0',
            'google-cloud-logging==2.7.0',
            'google-cloud-compute==0.7.0',
            'importlib-metadata==4.11.3',
            'protobuf==3.20.1',
            'click==8.1.3',
            'opencv-python==4.5.5.64',
            'google-api-python-client==1.9.3',
            'google-cloud-tasks==2.7.1',
            'requests==2.27.1',
            'numpy==1.22.3',
            'google-crc32c==1.3.0',
        ],
        packages=setuptools.find_packages(include=[
            '*ArgumentValidation*', '*DataStandardizer*',
            '*GoogleClientWrappers*', '*ClientTools*', '*ClientCommands*',
            'indxdatalaketools', '*DataLakeGlobals*', '*GoogleClients*',
            '*PropertiesDBApi*', '*PatientsHoldingApi*',
            '*DocumentClassifierApi*'
        ],
                                          exclude=[
                                              '**test**',
                                              '**pycache**',
                                              '*GoogleCloudResources*',
                                              '*ComputeEngineCreator*',
                                              '*ComputeEngineScripts*',
                                              '*DataLakeFileOperations*',
                                              '*DataLakeOperations*',
                                              '*AdminTools*',
                                          ]),
        python_requires=">=3.8",
        include_package_data=True,
        package_data={'': []},
    )



TIME_REGEX = "T(2[0-3]|[01][0-9])([0-5][0-9])([0-5][0-9])Z"
DATE_REGEX = "([1-2][0-9][0-9][0-9])(0[1-9]|1[0-2])(0[1-9]|[1-2][0-9]|3[01])"

SA_SUFFIX = '@indx-data-services.iam.gserviceaccount.com'

DATASET_OWNER = 'data-lake-generator@indx-data-services.iam.gserviceaccount.com'

SIZE_OF_FILES = 0.0
